<?php

namespace Ulmus\Ldap\Common;

use function ldap_set_option, ldap_start_tls, ldap_bind, ldap_unbind, ldap_connect, ldap_close, ldap_get_entries, ldap_mod_replace, ldap_count_entries, ldap_errno, ldap_error;

class LdapObject {

    public static ? string $dump = null;

    public $connection;

    public $search;

    public int $rowCount = 0;

    public int $bufferedRows = 0;

    public /*int*/ $lastInsertId = 0;

    public string $dn;

    public bool $binded = false;

    public function __destruct()
    {
        isset($this->connection) and ldap_close($this->connection);
    }

    public function connect(string $host, string $baseDn) : bool
    {
        $this->connection = ldap_connect("ldap://$host");

        $this->dn = $baseDn;

        return $this->connection !== false;
    }

    public function setOptions(array $options, $useConnectionRessource = true) : void
    {
        foreach($options as $field => $value) {
            ldap_set_option($useConnectionRessource ? $this->connection : null, $field, $value);
        }
    }

    public function startTLS() : void
    {
        ldap_start_tls($this->connection);
    }

    public function bind(? string $dn, ? string $password = null) : bool
    {
        if ($this->binded) {
            throw new \Exception("LdapObject is already binded with a user. Use the unbind() method to release it.");
        }

        $this->binded = ldap_bind($this->connection, $dn, $password);

        return $this->binded;
    }

    public function unbind() : void
    {
        if ($this->binded) {
            $this->binded = ! ldap_unbind($this->connection);
        }
    }

    public function rebind(? string $dn, ? string $password = null) : bool
    {
        $this->unbind();

        return $this->bind($dn, $password);
    }

    public function compare(? string $dn, string $attribute, /* mixed */ $value) /*: bool|int*/
    {
        return ldap_compare($this->connection, $dn, $attribute, $value);
    }

    public function select(array $filter, array $fields = [])
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $fields ]);

        $this->search = ldap_search($this->connection, $this->dn, $filter['filters'], $filter['fields'], 0, 0);

        $this->rowCount = $this->bufferedRows = ldap_count_entries($this->connection, $this->search);

        return $this;
    }

    public function fetch() /* : bool|array */
    {
        static $result;

        if (! $this->search ) {
            throw new \Exception('Impossible to fetch from LdapObject from which select() was not called first.');
        }

        if ( ! $this->bufferedRows ) {
            return $result = false;
        }
        if ( $this->rowCount === $this->bufferedRows ) {
            $result = ldap_first_entry($this->connection, $this->search);
        }
        else {
            $result = ldap_next_entry($this->connection, $result);
        }

        if ($result) {
            $this->bufferedRows--;

            $dataset = [
                'dn' => ldap_get_dn($this->connection, $result)
            ];

            if ( $attributes = ldap_get_attributes($this->connection, $result) ) {
                for ($i = 0; $i < $attributes['count']; $i++) {
                    $key = $attributes[$i];
                    $dataset[$key] = $attributes[$key]['count'] > 1 ? array_diff_key($attributes[$key], [ 'count' => "" ]) : $attributes[$key][0];
                }
            }

            return $dataset;
        }

        return false;
    }

    public function fetchAll()
    {
        return ldap_get_entries($this->connection, $result);
    }

    public function rowCount() : int
    {
        return $this->rowCount;
    }

    /**
     * @deprecated
     */
    public function runQuery(array $filter, array $dataset)
    {
        throw new \Exception("Method runQuery() cannot be run on this connector");
    }

    public function runInsertQuery(array $filter, array $dataset)
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $dataset ]);

        $dn = $filter['dn'] ?? $this->dn;

        $combine = array_combine($filter['fields'], $dataset);

        if ( $combine['cn'] ?? false ) {
            $dn = "cn={$combine['cn']}," . $dn;
        }

        try {
            if (false === ($queryResult = ldap_add($this->connection, $dn, $combine))) {
                $this->throwLdapException();
            }
        }
        catch(\Throwable $e) {
            $this->throwLdapException();
        }

        if ($queryResult) {
            $this->rowCount = (int)$queryResult;
            $this->lastInsertId = $dn;
        }

        return $this;
    }

    public function runUpdateQuery(array $filter, array $dataset)
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $dataset ]);
        
        if ( false === ( $queryResult = ldap_mod_replace($this->connection, $filter['dn'], $dataset) ) ) {
            $this->throwLdapException();
        }

        $this->rowCount = (int) $queryResult;

        return $this;
    }

    public function runDeleteQuery(array $filter, array $control)
    {
        static::$dump && call_user_func_array(static::$dump, [ $filter, $dataset ]);

        if ( false === ( $queryResult = ldap_delete($this->connection, $filter['dn']) ) ) {
            $this->throwLdapException();
        }

        $this->rowCount = (int) $queryResult;

        return $this;
    }

    protected function throwLdapException() : void
    {
        throw new \Exception(sprintf('LDAP error #%s `%s`', ldap_errno($this->connection), ldap_error($this->connection)));
    }

    public function closeCursor() : void {}
}
